import psycopg
import pytest
from psycopg import Connection
from pytest_postgresql import factories

from ..inconsistent_expression_checker import InconsistentExpressionChecker
from ..qepparser import QEPParser
from ..sqlparser import SqlParser

CUSTOMERS_TABLE_NAME = "inconsistent_expression_test_table_customers"
ORDERS_TABLE_NAME = "inconsistent_expression_test_table_orders"


def load_database(**kwargs):
    # pylint: disable=line-too-long
    conn: Connection = psycopg.connect(**kwargs)
    with conn.cursor() as cur:
        cur.execute(
            f"""
        DROP TABLE IF EXISTS {ORDERS_TABLE_NAME};
        DROP TABLE IF EXISTS {CUSTOMERS_TABLE_NAME};

        CREATE TABLE {CUSTOMERS_TABLE_NAME} (
        customer_id INT
        , fname VARCHAR(50) NOT NULL
        , sname VARCHAR(50) NOT NULL
        , type CHAR(1) NOT NULL
            CHECK (type IN ('C', 'B')) -- C = customer, B = business
        , nickname VARCHAR(20) NOT NULL
        , PRIMARY KEY (customer_id)
        );

        CREATE TABLE {ORDERS_TABLE_NAME} (
            order_id        INT
        , order_total_eur DECIMAL(6,2) NOT NULL
        , customer_id     INT           NOT NULL
        , PRIMARY KEY (order_id)
        , FOREIGN KEY (customer_id)
            REFERENCES {CUSTOMERS_TABLE_NAME} (customer_id)
        );

        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (1, 'Josi', 'Grimsell', 'B', 'Aaron');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (2, 'Tomlin', 'Nozzolinii', 'B', 'Abbigail');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (3, 'Christen', 'Culley', 'C', 'Abednego');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (4, 'Nancey', 'Fawlkes', 'C', 'Abel');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (5, 'Callida', 'Tomasello', 'B', 'Abiel');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (6, 'Daisey', 'Hamill', 'B', 'Abigail');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (7, 'Lorin', 'Dollimore', 'C', 'Abijah');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (8, 'Gasparo', 'Bohlje', 'C', 'Abner');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (9, 'Lester', 'Markus', 'C', 'Abraham');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (10, 'Sonnie', 'Kelling', 'B', 'Abram');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (11, 'Amby', 'Ligoe', 'C', 'Absalom');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (12, 'Tiffi', 'Riolfo', 'C', 'Ada');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (13, 'Marion', 'Penelli', 'B', 'Adaline');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (14, 'Davon', 'Burris', 'C', 'Addison');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (15, 'Zachary', 'Faloon', 'C', 'Adela');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (16, 'Augustin', 'Blaxall', 'C', 'Adelaide');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (17, 'Carlyle', 'Zimek', 'B', 'Adelbert');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (18, 'Odie', 'Rowling', 'C', 'Adele');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (19, 'Daphne', 'Bullen', 'B', 'Adeline');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (20, 'Robbi', 'O''Caherny', 'B', 'Adelphia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (21, 'Nady', 'Lempertz', 'B', 'Adolphus');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (22, 'Grazia', 'Syne', 'C', 'Adrian');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (23, 'Jeanette', 'Fincher', 'B', 'Adriane');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (24, 'Jacinthe', 'Kleeman', 'B', 'Adrienne');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (25, 'Neely', 'Merrydew', 'B', 'Agatha');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (26, 'Jenica', 'Martina', 'C', 'Agnes');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (27, 'Callean', 'Werlock', 'C', 'Aileen');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (28, 'Nanice', 'MacMaster', 'B', 'Alan');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (29, 'Dawna', 'Knipe', 'C', 'Alanson');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (30, 'Colin', 'Jansie', 'C', 'Alastair');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (31, 'Dario', 'Siehard', 'C', 'Alazama');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (32, 'Aubrie', 'Lockitt', 'C', 'Albert');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (33, 'Amble', 'Jewes', 'B', 'Alberta');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (34, 'Lorens', 'Buzin', 'C', 'Aldo');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (35, 'Lainey', 'Davidow', 'C', 'Aldrich');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (36, 'Bendite', 'Morfett', 'B', 'Aleksandr');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (37, 'Talbot', 'Keddey', 'B', 'Aleva');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (38, 'Web', 'Catterill', 'B', 'Alex');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (39, 'Tabbie', 'Glison', 'B', 'Alexander');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (40, 'Olva', 'Leborgne', 'C', 'Alexandra');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (41, 'Reilly', 'Kennler', 'B', 'Alexandria');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (42, 'Pyotr', 'Feldhorn', 'C', 'Alexis');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (43, 'Belle', 'Barsham', 'B', 'Alfonse');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (44, 'Buffy', 'O''Cridigan', 'B', 'Alfred');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (45, 'Seymour', 'Mayer', 'B', 'Alfreda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (46, 'Mechelle', 'Vinick', 'B', 'Algernon');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (47, 'Tracey', 'Sauvan', 'B', 'Alice');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (48, 'Marve', 'Eykel', 'C', 'Alicia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (49, 'Theodosia', 'Rosson', 'B', 'Aline');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (50, 'Anne-marie', 'Mounsie', 'B', 'Alison');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (51, 'Kennan', 'Burstowe', 'B', 'Alixandra');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (52, 'Gerrilee', 'Ackland', 'B', 'Allan');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (53, 'Riva', 'Scawen', 'C', 'Allen');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (54, 'Malchy', 'Hearty', 'B', 'Allisandra');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (55, 'Terrill', 'Syde', 'B', 'Allison');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (56, 'Kaitlin', 'Payle', 'B', 'Allyson');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (57, 'Vita', 'Dunnett', 'B', 'Allyssa');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (58, 'Lettie', 'Coffin', 'B', 'Almena');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (59, 'Lin', 'Race', 'B', 'Almina');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (60, 'Sibyl', 'Thoms', 'C', 'Almira');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (61, 'Blisse', 'Dillway', 'C', 'Alonzo');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (62, 'Seline', 'McGray', 'B', 'Alphinias');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (63, 'Gratia', 'Moss', 'B', 'Althea');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (64, 'Dyan', 'Crosbie', 'B', 'Alverta');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (65, 'Glenna', 'Alastair', 'C', 'Alyssa');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (66, 'Sadella', 'Boxhall', 'B', 'Alzada');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (67, 'Zsazsa', 'Bellocht', 'C', 'Amanda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (68, 'Leisha', 'Darlington', 'C', 'Ambrose');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (69, 'Maryjo', 'Pink', 'C', 'Amelia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (70, 'Carmela', 'Sedgeworth', 'C', 'Amos');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (71, 'Cynthie', 'Rouby', 'B', 'Anastasia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (72, 'Nadiya', 'Gingles', 'B', 'Anderson');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (73, 'Seymour', 'Maffioletti', 'B', 'Andre');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (74, 'Giffy', 'Cottee', 'C', 'Andrea');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (75, 'Erma', 'Cranstone', 'B', 'Andrew');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (76, 'Ardys', 'Currm', 'B', 'Andriane');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (77, 'Maegan', 'Wheldon', 'C', 'Angela');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (78, 'Matilde', 'Froome', 'B', 'Angelica');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (79, 'Preston', 'Groves', 'B', 'Angelina');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (80, 'Petey', 'Colloby', 'B', 'Ann');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (81, 'Sharline', 'Rosenwald', 'B', 'Anna');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (82, 'Abey', 'Darridon', 'C', 'Anne');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (83, 'Wolf', 'Brenton', 'B', 'Annette');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (84, 'Birk', 'Malling', 'C', 'Annie');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (85, 'Brig', 'Triswell', 'C', 'Anselm');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (86, 'Lauren', 'Tolworth', 'B', 'Anthony');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (87, 'Nanette', 'McElwee', 'B', 'Antoinette');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (88, 'Sansone', 'Copsey', 'C', 'Antonia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (89, 'Nickie', 'Bloss', 'C', 'Antonio');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (90, 'Forest', 'Trim', 'B', 'Appoline');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (91, 'Bobby', 'Fortesquieu', 'B', 'Aquilla');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (92, 'Raven', 'Eilles', 'C', 'Ara');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (93, 'Sashenka', 'Fedorski', 'C', 'Arabella');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (94, 'Halimeda', 'Freak', 'B', 'Arabelle');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (95, 'Amandie', 'Botham', 'B', 'Araminta');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (96, 'Tiertza', 'Bunford', 'C', 'Archibald');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (97, 'Germain', 'Haly', 'B', 'Archilles');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (98, 'Elfreda', 'Tome', 'C', 'Ariadne');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (99, 'Phineas', 'Yuryaev', 'B', 'Arielle');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (100, 'Gypsy', 'Bottrell', 'C', 'Aristotle');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (101, 'Hamilton', 'Pellew', 'B', 'Arizona');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (102, 'Ricca', 'Rupprecht', 'B', 'Arlene');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (103, 'Ede', 'Golden of Ireland', 'C', 'Armanda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (104, 'Timofei', 'Grene', 'B', 'Armena');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (105, 'Cary', 'Sells', 'C', 'Armilda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (106, 'Merrily', 'Coundley', 'C', 'Arminda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (107, 'Celisse', 'Rubra', 'B', 'Arminta');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (108, 'Janene', 'Motten', 'C', 'Arnold');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (109, 'Marquita', 'Impy', 'B', 'Aron');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (110, 'Donall', 'Labrenz', 'B', 'Artelepsa');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (111, 'Torey', 'Sessuns', 'B', 'Artemus');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (112, 'Nanci', 'Byrd', 'C', 'Arthur');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (113, 'Dania', 'Foxton', 'C', 'Arthusa');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (114, 'Fara', 'Arkil', 'B', 'Arzada');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (115, 'Shirl', 'Pilpovic', 'C', 'Asahel');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (116, 'Annmarie', 'Errichelli', 'B', 'Asaph');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (117, 'Marysa', 'Wadman', 'C', 'Asenath');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (118, 'Hermy', 'Whieldon', 'B', 'Ashley');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (119, 'Gregg', 'Lazenbury', 'B', 'Aubrey');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (120, 'Corinna', 'Pyburn', 'B', 'Audrey');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (121, 'Clywd', 'Stokey', 'B', 'August');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (122, 'Myrlene', 'Worcs', 'B', 'Augusta');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (123, 'Lyon', 'Nolleau', 'C', 'Augustina');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (124, 'Penelope', 'Quadling', 'C', 'Augustine');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (125, 'Leo', 'Creaney', 'B', 'Augustus');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (126, 'Lilllie', 'Annis', 'C', 'Aurelia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (127, 'Rhoda', 'Kurton', 'C', 'Avarilla');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (128, 'Jere', 'Cometti', 'B', 'Azariah');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (129, 'Ashla', 'De-Ville', 'C', 'Bab');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (130, 'Rora', 'Adcock', 'C', 'Babs');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (131, 'Ive', 'Matches', 'B', 'Barbara');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (132, 'Olympe', 'Faber', 'B', 'Barbery');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (133, 'Clementia', 'Fergusson', 'B', 'Barbie');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (134, 'Georgetta', 'Crossley', 'B', 'Barnabas');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (135, 'Janenna', 'McCamish', 'C', 'Barney');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (136, 'Aretha', 'Arias', 'C', 'Bart');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (137, 'Lucilia', 'Bentick', 'C', 'Bartholomew');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (138, 'Ainsley', 'Wraighte', 'C', 'Barticus');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (139, 'Kelsi', 'Suddell', 'C', 'Bazaleel');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (140, 'Suki', 'Flinn', 'B', 'Bea');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (141, 'Jacinta', 'Villiers', 'B', 'Beatrice');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (142, 'Elfrieda', 'Cleary', 'C', 'Becca');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (143, 'Sasha', 'Eunson', 'C', 'Beck');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (144, 'Ange', 'Pasterfield', 'C', 'Bedelia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (145, 'Adena', 'Wenham', 'C', 'Belinda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (146, 'Dunstan', 'Cressey', 'B', 'Bella');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (147, 'Wilmer', 'MacWhan', 'B', 'Benedict');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (148, 'Lidia', 'Quinane', 'C', 'Benjamin');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (149, 'Erik', 'Cushe', 'B', 'Benjy');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (150, 'Agnesse', 'Liebmann', 'B', 'Bernard');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (151, 'Prentiss', 'Filby', 'B', 'Berney');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (152, 'Towny', 'Middell', 'B', 'Bert');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (153, 'Alfy', 'McCahill', 'B', 'Bertha');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (154, 'Conney', 'Riteley', 'C', 'Bertram');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (155, 'Murdock', 'Dix', 'C', 'Bess');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (156, 'Allistir', 'Frary', 'C', 'Beth');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (157, 'Noach', 'MacRitchie', 'C', 'Bethena');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (158, 'Anett', 'Hercules', 'B', 'Beverly');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (159, 'Carilyn', 'Vinecombe', 'B', 'Bezaleel');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (160, 'Patti', 'Louis', 'C', 'Biddie');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (161, 'Barron', 'Dishmon', 'B', 'Bill');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (162, 'Jessee', 'List', 'B', 'Billy');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (163, 'Gerhardt', 'Simcock', 'B', 'Blanche');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (164, 'Abel', 'Bezants', 'C', 'Bob');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (165, 'Angele', 'Wildbore', 'C', 'Bobby');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (166, 'Angelo', 'Vanstone', 'C', 'Boetius');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (167, 'Ninnette', 'Steere', 'C', 'Brad');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (168, 'Melicent', 'Cheston', 'B', 'Bradford');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (169, 'Elsbeth', 'Nucciotti', 'B', 'Bradley');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (170, 'Bard', 'Shaw', 'C', 'Brady');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (171, 'Fredrick', 'Matuszynski', 'C', 'Breanna');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (172, 'Pammi', 'Codling', 'B', 'Breeanna');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (173, 'Kristopher', 'Mackelworth', 'B', 'Brenda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (174, 'Carolyn', 'Quinlan', 'C', 'Brian');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (175, 'Davy', 'Blomefield', 'B', 'Brianna');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (176, 'Marylou', 'Guilloton', 'C', 'Bridget');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (177, 'Genvieve', 'Marthen', 'B', 'Brittany');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (178, 'Rogers', 'Plose', 'B', 'Brittney');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (179, 'Daphne', 'Kells', 'C', 'Broderick');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (180, 'Leonelle', 'Chiommienti', 'B', 'Caitlin');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (181, 'Daria', 'Wistance', 'C', 'Caitlyn');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (182, 'Sasha', 'Stiffkins', 'C', 'Caldonia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (183, 'Harvey', 'Skinn', 'B', 'Caleb');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (184, 'Louise', 'Jansey', 'C', 'California');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (185, 'Tanitansy', 'Headington', 'C', 'Calista');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (186, 'Chloette', 'Ratt', 'B', 'Calpurnia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (187, 'Donall', 'Casolla', 'B', 'Calvin');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (188, 'Ferguson', 'Mackriell', 'B', 'Cameron');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (189, 'Merle', 'Erridge', 'C', 'Camille');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (190, 'Kara', 'Backs', 'B', 'Campbell');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (191, 'Frankie', 'Canny', 'B', 'Candace');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (192, 'Dudley', 'Vesco', 'B', 'Carlotta');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (193, 'Brucie', 'Coning', 'C', 'Carlton');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (194, 'Corilla', 'Aldwinckle', 'C', 'Carmellia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (195, 'Dewey', 'Bottom', 'C', 'Carmelo');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (196, 'Louisa', 'Matasov', 'B', 'Carmon');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (197, 'Missie', 'Fealty', 'C', 'Carol');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (198, 'Kellby', 'Threlfall', 'C', 'Carolann');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (199, 'Matilda', 'Malenoir', 'C', 'Caroline');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (200, 'Fanni', 'Iacopetti', 'C', 'Carolyn');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (201, 'Arv', 'Spawforth', 'B', 'Carrie');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (202, 'Sonja', 'Pentycross', 'B', 'Carthaette');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (203, 'Rory', 'Hallatt', 'C', 'Casey');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (204, 'Cassy', 'Pearcehouse', 'C', 'Casper');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (205, 'Orlan', 'Kynder', 'C', 'Cassandra');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (206, 'Ardyth', 'Conningham', 'B', 'Cassidy');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (207, 'Elise', 'Kamenar', 'C', 'Caswell');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (208, 'Legra', 'Drought', 'C', 'Catherine');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (209, 'Dulcea', 'Akister', 'C', 'Cathleen');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (210, 'Pamelina', 'Vittery', 'C', 'Cathy');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (211, 'Jobi', 'Bails', 'B', 'Cecilia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (212, 'Annecorinne', 'Soles', 'B', 'Cedric');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (213, 'Rubin', 'Lunney', 'B', 'Celeste');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (214, 'Chalmers', 'Britton', 'C', 'Celinda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (215, 'Nollie', 'Jemmett', 'B', 'Charity');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (216, 'Elsbeth', 'MacGaughie', 'B', 'Charles');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (217, 'Lucy', 'Durtnell', 'C', 'Charlie');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (218, 'Raffarty', 'Sweeney', 'C', 'Charlotte');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (219, 'Adriane', 'Ccomini', 'B', 'Chauncey');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (220, 'Kylen', 'Bewicke', 'B', 'Cheryl');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (221, 'Cassondra', 'Mattingson', 'B', 'Chesley');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (222, 'Hazel', 'Rushforth', 'C', 'Chester');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (223, 'Annice', 'Haynes', 'B', 'Chet');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (224, 'Dayle', 'Eirwin', 'B', 'Chick');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (225, 'Egbert', 'Vasyunin', 'B', 'Chloe');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (226, 'Emelina', 'Ayliff', 'B', 'Chris');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (227, 'Jeralee', 'Atwill', 'C', 'Christa');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (228, 'Rudd', 'Ames', 'B', 'Christian');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (229, 'Melessa', 'MacAleese', 'C', 'Christiana');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (230, 'Cordell', 'Bineham', 'B', 'Christiano');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (231, 'Calhoun', 'BoHlingolsen', 'C', 'Christina');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (232, 'Maritsa', 'Cowpland', 'C', 'Christine');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (233, 'Tibold', 'Sleigh', 'C', 'Christoph');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (234, 'Cayla', 'Statter', 'B', 'Christopher');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (235, 'Justus', 'Barthrup', 'B', 'Christy');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (236, 'Garnette', 'Tabrett', 'C', 'Cicely');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (237, 'Jeramie', 'Gallehock', 'C', 'Cinderella');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (238, 'Gregor', 'Crinion', 'B', 'Cindy');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (239, 'Atalanta', 'Girdler', 'C', 'Claire');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (240, 'Britte', 'Muge', 'B', 'Clara');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (241, 'Sigismund', 'Mowsdill', 'B', 'Clare');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (242, 'Brenden', 'Simmons', 'B', 'Clarence');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (243, 'Hannah', 'Mattheis', 'C', 'Clarinda');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (244, 'Katrine', 'Janiak', 'B', 'Clarissa');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (245, 'Holly', 'Roy', 'C', 'Claudia');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (246, 'Albert', 'Heimann', 'B', 'Cleatus');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (247, 'Dulcie', 'Crutchley', 'B', 'Clement');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (248, 'Norby', 'Biernacki', 'C', 'Clementine');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (249, 'Rhoda', 'Leheude', 'C', 'Cliff');
        insert into {CUSTOMERS_TABLE_NAME} (customer_id, fname, sname, type, nickname) values (250, 'Langston', 'Prosser', 'C', 'Clifford');

        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (1, 535.36, 111);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (2, 409.8, 217);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (3, 189.43, 19);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (4, 144.14, 157);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (5, 582.52, 172);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (6, 132.85, 206);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (7, 183.92, 236);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (8, 424.8, 244);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (9, 519.43, 175);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (10, 414.55, 234);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (11, 88.19, 50);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (12, 591.72, 143);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (13, 503.52, 216);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (14, 586.06, 181);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (15, 47.79, 248);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (16, 330.92, 130);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (17, 302.31, 225);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (18, 438.38, 26);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (19, 107.53, 94);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (20, 207.6, 9);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (21, 471.12, 179);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (22, 193.12, 6);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (23, 236.48, 51);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (24, 538.88, 38);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (25, 83.54, 79);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (26, 590.52, 50);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (27, 137.86, 21);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (28, 87.44, 1);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (29, 217.18, 124);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (30, 435.57, 105);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (31, 394.48, 62);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (32, 331.93, 144);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (33, 140.92, 236);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (34, 64.76, 13);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (35, 389.81, 218);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (36, 225.6, 136);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (37, 322.11, 41);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (38, 366.31, 59);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (39, 42.62, 95);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (40, 33.89, 23);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (41, 158.78, 145);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (42, 561.85, 222);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (43, 133.73, 196);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (44, 537.56, 2);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (45, 46.05, 177);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (46, 362.72, 240);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (47, 163.79, 141);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (48, 199.52, 234);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (49, 376.9, 89);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (50, 306.71, 235);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (51, 244.27, 119);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (52, 81.47, 212);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (53, 592.15, 207);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (54, 591.44, 145);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (55, 391.93, 128);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (56, 501.06, 129);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (57, 349.14, 58);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (58, 10.24, 74);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (59, 141.8, 118);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (60, 473.05, 208);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (61, 343.91, 223);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (62, 418.98, 60);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (63, 542.48, 52);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (64, 534.3, 108);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (65, 318.64, 6);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (66, 454.41, 11);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (67, 550.21, 63);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (68, 73.27, 189);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (69, 266.3, 39);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (70, 596.2, 221);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (71, 175.29, 115);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (72, 539.87, 62);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (73, 144.33, 52);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (74, 63.02, 117);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (75, 238.65, 203);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (76, 281.92, 183);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (77, 335.23, 149);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (78, 569.4, 99);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (79, 94.87, 194);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (80, 17.1, 21);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (81, 490.65, 41);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (82, 106.39, 237);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (83, 451.45, 203);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (84, 396.02, 111);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (85, 404.96, 79);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (86, 537.64, 201);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (87, 441.16, 191);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (88, 210.95, 158);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (89, 556.68, 74);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (90, 169.25, 30);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (91, 331.76, 115);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (92, 156.72, 192);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (93, 452.63, 115);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (94, 406.59, 129);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (95, 367.31, 48);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (96, 311.78, 173);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (97, 313.39, 241);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (98, 177.22, 86);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (99, 354.37, 66);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (100, 424.83, 50);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (101, 328.66, 189);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (102, 158.53, 220);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (103, 14.83, 183);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (104, 97.04, 11);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (105, 107.29, 12);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (106, 236.57, 4);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (107, 23.63, 179);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (108, 294.25, 131);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (109, 170.82, 206);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (110, 144.28, 59);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (111, 490.85, 191);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (112, 38.97, 239);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (113, 305.06, 123);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (114, 339.9, 108);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (115, 351.68, 153);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (116, 292.86, 97);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (117, 23.13, 46);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (118, 23.96, 195);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (119, 579.12, 43);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (120, 454.18, 214);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (121, 132.25, 50);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (122, 406.1, 80);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (123, 370.87, 115);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (124, 552.02, 191);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (125, 170.66, 131);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (126, 384.61, 226);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (127, 451.77, 17);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (128, 415.86, 243);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (129, 426.3, 136);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (130, 575.58, 235);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (131, 255.66, 127);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (132, 488.95, 234);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (133, 499.73, 235);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (134, 296.72, 222);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (135, 574.1, 222);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (136, 411.87, 10);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (137, 417.16, 215);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (138, 223.81, 183);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (139, 441.16, 97);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (140, 297.41, 101);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (141, 183.27, 44);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (142, 509.25, 190);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (143, 239.98, 109);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (144, 504.73, 154);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (145, 112.51, 162);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (146, 184.13, 22);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (147, 180.39, 2);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (148, 203.36, 180);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (149, 340.62, 215);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (150, 439.75, 18);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (151, 221.16, 212);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (152, 291.54, 10);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (153, 350.78, 247);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (154, 291.52, 54);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (155, 274.14, 151);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (156, 130.63, 188);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (157, 477.53, 205);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (158, 65.82, 201);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (159, 281.36, 198);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (160, 144.6, 82);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (161, 184.39, 94);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (162, 168.92, 27);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (163, 61.42, 69);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (164, 193.39, 169);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (165, 84.61, 1);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (166, 95.01, 189);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (167, 547.07, 197);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (168, 137.02, 109);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (169, 552.88, 241);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (170, 539.11, 92);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (171, 175.72, 249);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (172, 15.3, 30);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (173, 297.4, 85);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (174, 597.34, 184);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (175, 161.61, 58);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (176, 22.8, 175);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (177, 334.0, 167);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (178, 563.08, 161);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (179, 453.95, 43);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (180, 452.67, 234);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (181, 150.37, 69);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (182, 478.47, 70);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (183, 546.08, 88);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (184, 380.46, 98);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (185, 505.7, 49);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (186, 561.85, 156);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (187, 271.52, 242);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (188, 495.68, 39);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (189, 164.91, 68);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (190, 401.19, 88);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (191, 183.55, 142);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (192, 206.39, 151);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (193, 449.91, 213);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (194, 328.76, 41);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (195, 504.28, 117);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (196, 259.0, 244);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (197, 529.63, 46);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (198, 120.47, 95);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (199, 376.5, 53);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (200, 592.16, 137);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (201, 411.07, 202);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (202, 99.52, 171);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (203, 545.33, 116);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (204, 212.04, 203);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (205, 508.59, 59);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (206, 226.62, 228);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (207, 205.25, 199);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (208, 344.91, 150);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (209, 203.86, 244);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (210, 332.45, 47);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (211, 399.2, 138);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (212, 179.55, 180);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (213, 329.78, 105);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (214, 520.52, 98);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (215, 515.79, 96);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (216, 388.15, 142);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (217, 302.24, 189);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (218, 389.51, 177);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (219, 594.84, 121);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (220, 367.15, 163);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (221, 487.39, 227);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (222, 416.65, 59);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (223, 200.44, 47);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (224, 293.35, 186);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (225, 26.43, 228);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (226, 212.04, 28);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (227, 189.25, 230);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (228, 542.3, 42);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (229, 89.34, 147);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (230, 389.39, 247);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (231, 51.77, 118);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (232, 330.85, 204);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (233, 87.21, 179);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (234, 254.56, 98);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (235, 230.72, 168);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (236, 485.18, 167);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (237, 293.23, 140);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (238, 448.86, 16);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (239, 327.06, 34);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (240, 384.87, 107);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (241, 296.03, 169);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (242, 356.69, 244);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (243, 209.34, 31);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (244, 273.35, 153);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (245, 327.9, 212);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (246, 510.74, 188);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (247, 123.55, 179);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (248, 321.97, 195);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (249, 491.05, 63);
        insert into {ORDERS_TABLE_NAME} (order_id, order_total_eur, customer_id) values (250, 367.56, 214);"""
        )
        conn.commit()


factory = factories.postgresql_proc(
    load=[load_database],
)
postgresql = factories.postgresql("factory")


@pytest.fixture
def sql_parser(postgresql: Connection):
    return SqlParser(db_connection=postgresql)


@pytest.fixture
def qep_parser(postgresql: Connection):
    return QEPParser(conn=postgresql)


def test_check(sql_parser: SqlParser, qep_parser: QEPParser):
    SQL_SIMPLE = """
SELECT (1, 2, 3);"""

    SQL_INCONSISTENT_CMP = f"""
SELECT order_id
FROM {ORDERS_TABLE_NAME}
WHERE order_total_eur = 10 AND order_total_eur = 50;"""

    sql_statement = SQL_SIMPLE
    parsed_sql = sql_parser.parse_one(sql_statement)
    qep_analysis = qep_parser.parse(sql_statement)
    checker = InconsistentExpressionChecker(parsed_sql, qep_analysis)
    assert checker is not None
    warning_msg = checker.check()
    assert warning_msg is None

    sql_statement = SQL_INCONSISTENT_CMP
    parsed_sql = sql_parser.parse_one(sql_statement)
    qep_analysis = qep_parser.parse(sql_statement)
    checker = InconsistentExpressionChecker(parsed_sql, qep_analysis)
    assert checker is not None
    warning_msg = checker.check()
    assert warning_msg is not None



import pexpect
from typing import Optional, Tuple

from pyparsing import Literal, Word, ParseException, ParserElement, \
    identbodychars, nums


class PsqlConnInfo:
    """Get PostgreSQL server address, port, database name, and user via \
    psql by supplying same arguments as the original psql process. This \
    way we can avoid writing a command-line argument parser."""

    # tok, or token, is parsing element with only single element output,
    # either by only having single element, or using Combine to squash
    # multiple elements into one. These are often combined to build functions
    # for matching.

    tok_pre_database: ParserElement = \
        Literal("You are connected to database \"")
    tok_database: ParserElement = \
        Word(identbodychars)

    tok_pre_user: ParserElement = \
        Literal("\" as user \"")
    tok_user: ParserElement = \
        Word(identbodychars)

    tok_pre_host: ParserElement = \
        Literal("\" on host \"") | \
        Literal("\" via socket in \"")
    tok_host: ParserElement = \
        Word(identbodychars + "/.")

    tok_pre_port: ParserElement = \
        Literal("\" at port \"")
    tok_port: ParserElement = \
        Word(nums)
    tok_end: ParserElement = \
        Literal("\".")

    def __init__(
        self,
        psql_args: str
    ):
        """Use psql child process with exact same command-line arguments \
        to initialize connection info."""
        self.pg_tuple: Optional[Tuple[str, str, str, str, str]] = None
        match_psql_conninfo: ParserElement = (
            self.tok_pre_database + self.tok_database +
            self.tok_pre_user + self.tok_user +
            self.tok_pre_host + self.tok_host +
            self.tok_pre_port + self.tok_port +
            self.tok_end
        )
        pexpect_conninfo = pexpect.spawn("psql -c \"\\conninfo\" " + psql_args)
        pexpect_conninfo.expect(pexpect.EOF)
        conninfo_str = bytes.decode(pexpect_conninfo.before)
        try:
            conninfo_res = \
                match_psql_conninfo.parse_string(conninfo_str).as_list()
        except ParseException:
            return
        self.pg_host = conninfo_res[5]
        self.pg_port = conninfo_res[7]
        self.pg_user = conninfo_res[3]
        self.pg_pass = ""
        self.pg_name = conninfo_res[1]
        self.pg_tuple = (
            self.pg_host,
            self.pg_port,
            self.pg_user,
            self.pg_pass,
            self.pg_name
        )
        pass

    def get(
            self
    ) -> Optional[Tuple[str, str, str, str, str]]:
        """Get 5-tuple that has the PostgreSQL host, port, user, pass, and db \
        name.

        :returns: (hostname, port, user, password, database name) \
        5-tuple or None.
        """
        return self.pg_tuple



import re
import sys
from dataclasses import dataclass
from typing import Optional, TextIO

from .config_values import ConfigValues


class ConfigParser:
    _option_matcher: re.Pattern = re.compile(
        r"\s*(?P<optname>\w+)\s+(?P<optval>1|0|true|false|yes|no)\s*$",
        flags=re.IGNORECASE,
    )
    _empty_line_matcher: re.Pattern = re.compile(r"^\s*$")
    _comment_matcher: re.Pattern = re.compile(r"^\s*#+.*$")

    def __init__(self, file: TextIO):
        self.file: TextIO = file

    def parse(self) -> Optional[ConfigValues]:
        """
        Reads config values from file givein in __init__.
        """

        optnames = [x.lower() for x in ConfigValues.__annotations__.keys()]
        config_values: ConfigValues = {}

        # Needed for bytes containing files
        self.file.seek(0)

        # The following stuff is for improved error messages
        @dataclass(frozen=True)
        class SeenOptionContext:
            key: str
            line: str
            line_number: int

        seen_option_contexts: list[SeenOptionContext] = []
        multiply_defined_options = set()

        for line_number, line in enumerate(self.file.readlines()):
            # To make this work with bytes objects like TemporaryFile contents
            if isinstance(line, bytes):
                line = bytes.decode(line, "utf-8")
            line = str(line)

            if match := ConfigParser._empty_line_matcher.match(line):
                continue

            if match := ConfigParser._comment_matcher.match(line):
                continue

            if match := ConfigParser._option_matcher.match(line):
                optname = match.group("optname")
                if optname.lower() in optnames:
                    key = self._convert_from_anycase_to_propercase(optname)
                    config_values[key] = self._optval_to_bool(
                        str(match.group("optval"))
                    )

                    if key in [x.key for x in seen_option_contexts]:
                        seen_option_contexts.append(
                            SeenOptionContext(
                                key=key, line=line.rstrip("\n"), line_number=line_number
                            )
                        )
                        multiply_defined_options.add(key)
                    else:
                        seen_option_contexts.append(
                            SeenOptionContext(
                                key=key, line=line.rstrip("\n"), line_number=line_number
                            )
                        )
                else:
                    output_line = line.rstrip("\n")
                    print(
                        f"warning: bad warning name or value in line {line_number}: '{output_line}' in configuration file: '{self.file.name}'",
                        file=sys.stderr,
                    )
            else:
                output_line = line.rstrip("\n")
                print(
                    f"warning: unable to parse line {line_number}: '{output_line}' in configuration file: '{self.file.name}'",
                    file=sys.stderr,
                )

        # Not so brilliant computational complexity
        for key in multiply_defined_options:
            seen_option_lines = ""
            for ctx in filter(lambda x: x.key == key, seen_option_contexts):
                seen_option_lines += f"line {ctx.line_number}: '{ctx.line}'\n"
            print(
                f"warning: option '{key}' is set multiple times\n{seen_option_lines}in configuration file: '{self.file.name}'",
                file=sys.stderr,
            )

        return config_values if len(config_values) > 0 else None

    def _optval_to_bool(self, optval: str) -> bool:
        """
        Excepts only valid option values.
        Converts config file option value string into bool.
        """

        true_values = ("true", "1", "yes")
        false_values = ("false", "0", "no")

        if optval.lower() in true_values:
            return True
        if optval.lower() in false_values:
            return False

        return False

    def _convert_from_anycase_to_propercase(self, anycase_key: str) -> str:
        """
        Assumes that anycase_key matches some option name case-insensitively.

        Users can write option values case-insensitively. We still need to
        convert that user written value to the proper one that matches the
        fields in ConfigValues class.
        """

        fields = ConfigValues.__annotations__.keys()
        for field in fields:
            if anycase_key.lower() == field.lower():
                return field



from typing import Optional

import sqlglot.expressions as exp

from .qepparser import QEPAnalysis
from .errfmt import ErrorFormatter


class SubqueryOrderByChecker:
    def __init__(self, parsed_sql: exp.Expression, qep_analysis: QEPAnalysis):
        self.parsed_sql: exp.Expression = parsed_sql
        self.qep_analysis: QEPAnalysis = qep_analysis

    def check(self) -> Optional[str]:
        """
        Returns warning message if there exists ORDER BY in a subquery,
        otherwise None.

        This check gives misses some situations with redundant ORDER BY but
        should never give false positives, only false negatives.
        """
        # TODO: More sophisticated check that inspects self.parsed_sql and
        #       finds more warnings than postgresql.

        if self.qep_analysis is None:
            return None

        has_orderby = self.parsed_sql.find(exp.Order) is not None
        has_sort_node = len(
            self.qep_analysis.root.rfindval("Node Type", "Sort")) > 0
        has_inner_orderby = has_orderby and not has_sort_node

        if not has_inner_orderby:
            return None

        warning = "ORDER BY in a subquery"
        warning_name = type(self).__name__.rstrip("Checker")
        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg



# Written by Tatu Heikkilä, tatu.heikkila@tuni.fi
# Licensed under MIT.
"""Interface with psql and capture all input and output.

Uses pexpect in combination with pyte for interfacing and screen-scraping
respectively.
"""

from copy import deepcopy
from shutil import get_terminal_size
from typing import Callable, List

import pexpect
from pyte import Stream, Screen

from .psqlparser import PsqlParser


class PsqlWrapper:
    """Handles terminal interfacing with psql, using the parameter parser \
    to pick up relevant SQL statements and syntax errors for hook functions."""

    # debug creates pyte.screen (current screenscraping context) and
    # psqlwrapper.log (capturing terminal stream) in working directory
    debug: bool = False

    supported_psql_versions: list[str] = ["14.5"]

    def __init__(
        self,
        psql_args: bytes,
        hook_semantic_f: Callable[[str], str],
        hook_syntax_f: Callable[[str], str],
        parser: PsqlParser
    ):
        """Build wrapper for selected database.

        :param psql_args: are the command-line arguments pg4n has been called \
        with.
        :param hook_semantic_f: is a callback to which scraped SQL queries are\
        passed to, and from which corresponding semantic warning messages are \
        received in return.
        :param hook_syntax_f: is a callback to which scraped syntax error \
        messages are passed to, and from which corresponding warning messages \
        are received.
        :param parser: A parser that implements the required parsing functions.
        """
        self.psql_args: bytes = psql_args
        self.semantic_analyze: Callable[[str], str] = hook_semantic_f
        self.syntax_analyze: Callable[[str], str] = hook_syntax_f
        self.parser: PsqlParser = parser

        # shutil.get_terminal_size()
        (self.cols, self.rows) = get_terminal_size()

        # pyte.Screen, pyte.Stream
        self.pyte_screen: Screen = Screen(self.cols, self.rows)
        self.pyte_screen_output_sink: Stream = Stream(self.pyte_screen)

        # Semantic analysis is always done when user presses Return
        # and resulting message is saved here until when new prompt comes in
        self.pg4n_message: str = ""

    def start(
        self
    ) -> None:
        """Start psql process and feed hook functions with \
        intercepted queries and syntax errors.

        Control is only returned after psql process exits.
        """
        version_msg = self._check_psql_version()
        if version_msg != "":
            print(version_msg)

        c = pexpect.spawn(
            "psql " + bytes.decode(self.psql_args),
            encoding="utf-8",
            dimensions=(self.rows, self.cols)
        )

        c.interact(input_filter=lambda x: x, output_filter=self._intercept)

    def _check_psql_version(self) -> str:
        """Check PostgreSQL version via psql child process and match \
        against versions pg4n is tested with.

        :returns: an empty string if current version has been tested, \
        otherwise a warning message.
        """
        version_info = pexpect.spawn(
            "psql " + bytes.decode(self.psql_args) + " --version"
        )
        version_info.expect(pexpect.EOF)
        version_info_str: str = bytes.decode(version_info.before)
        version: str = self.parser.parse_psql_version(version_info_str)

        # command-line arguments prevent interactive sessions (e.g pg4n --help)
        if version == "":
            return ""

        version_ok: bool = version in self.supported_psql_versions
        if version_ok:
            return ""
        else:
            return (
                "Pg4n has only been tested on psql versions "
                + str(self.supported_psql_versions)
                + "."
            )

    def _intercept(
        self,
        output: bytes
    ) -> bytes:
        """Forward output to `_check_and_act_on_repl_output` and feed \
        output to pyte screen for screenscraping.

        :param output: output seen on terminal screen.
        :returns: output with injected semantic error messages.
        """
        new_output: bytes = self._check_and_act_on_repl_output(output)

        self.pyte_screen_output_sink.feed(bytes.decode(new_output))

        if self.debug:
            f = open("pyte.screen", "w")
            f.write(
                '\n'.join(line.rstrip() for line in self.pyte_screen.display)
            )
            f.close()
            g = open("psqlwrapper.log", "a")
            g.write(str(new_output) + '\n')
            g.close()

        return new_output

    def _check_and_act_on_repl_output(
        self,
        latest_output: bytes
    ) -> bytes:
        """Check if user has hit Return so we can start analyzing, \
        or if a fresh prompt has come in and we can show them a helpful \
        message.

        :param latest_output: is used for Return press and fresh prompt \
        analysis. It is also what the helpful message will be injected to.
        :returns: output with injected semantic error messages.
        """
        new_output: bytes = ""
        # for optimization reasons, check output only if len() > 1, so most
        # keyboard input does not trigger parsing
        # (Return is always at least 2 length)
        if len(latest_output) <= 1:
            return latest_output

        # User hit Return: parse for potential SQL query, analyze, and
        # save a potential warning to be included in before next fresh prompt.
        if self._user_hit_return(latest_output):
            # get terminal screen contents
            screen: str = \
                '\n'.join(line.rstrip() for line in self.pyte_screen.display)

            parsed_sql_query: str = self.parser.parse_last_stmt(screen)
            if parsed_sql_query != "":
                # feed query to semantic analysis hook function
                # and save resulting message
                self.pg4n_message = self.semantic_analyze(parsed_sql_query)

        # If there is a fresh prompt:
        if self.parser.output_has_new_prompt(
                bytes.decode(latest_output)
        ):
            # If we have a semantic error message waiting
            if self.pg4n_message != "":
                new_output = self._replace_prompt(latest_output)
                self.pg4n_message = ""
                return new_output

            # Since latest_output contains error details, we will have to
            # see how the screen would look like, but still allow injecting
            # an insightful syntax error message from the syntax analysis.
            potential_future_screen = \
                deepcopy(self.pyte_screen)
            potential_future_screen_output_sink = \
                Stream(potential_future_screen)
            potential_future_screen_output_sink.feed(
                bytes.decode(latest_output)
            )

            potential_future_contents: str = '\n'.join(
                line.rstrip() for line in potential_future_screen.display
            )
            syntax_error = \
                self.parser.parse_syntax_error(potential_future_contents)
            if syntax_error != "":
                self.pg4n_message = self.syntax_analyze(syntax_error)
                new_output = self._replace_prompt(latest_output)
                self.pg4n_message = ""
                return new_output

        return latest_output

    def _replace_prompt(self, prompt: bytes) -> bytes:
        """Inject saved semantic error message into given prompt.

        :param prompt: is output where message is injected to. A fresh prompt \
        is expected, or otherwise no injection is made.
        :returns: a prompt with injected message, or unchanged if \
        detecting new prompt fails.
        """
        split_prompt: List[str] = self.parser.parse_new_prompt_and_rest(
                bytes.decode(prompt, "utf-8")
        )
        if split_prompt == []:
            return prompt  # prompt is malformed and is returned as-is.
        print_msg = self.pg4n_message.replace("\n", "\r\n")
        return bytes(
            split_prompt[0] + "\r\n"
            + print_msg + "\r\n\r\n"
            + split_prompt[1],
            "utf-8"
        )

    def _user_hit_return(self, output: bytes) -> bool:
        """Check if user hit return.

        :param output: is raw console output to be checked for return press.
        :returns: if user has hit return.
        """
        # trivial case:
        # we assume only situation with \r\n at start of
        # line is when user has pressed enter
        if output[0:2] == b"\r\n":
            return True

        # complicated case: user has ctrl-R'd, copypasted command or something.
        # and the \r\n is somewhere in midst of output..
        if self.parser.output_has_magical_return(
                bytes.decode(output, "utf-8")
        ):
            return True
        return False



# Written by Tatu Heikkilä, tatu.heikkila@tuni.fi
# Licensed under MIT.
"""Parse psql output."""

from functools import reduce
from string import printable
from typing import Optional

from pyparsing import (
    CaselessLiteral,
    Char,
    Combine,
    Literal,
    Opt,
    ParseException,
    ParseResults,
    ParserElement,
    StringEnd,
    White,
    Word,
    ZeroOrMore,
    identbodychars,
    nums
)

# Turn on memoization optimization
ParserElement.enablePackrat(None)

# default whitespace rules complicate things needlessly, remove them:
ParserElement.setDefaultWhitespaceChars("")


class PsqlParser:
    """Parses psql output for syntactic analysis."""

    # Turn on verbose output to psqlparser.log file in working directory
    debug: bool = False

    # Parsing functions common to more than 1 parsing functions are listed here

    prompt_chars: str = identbodychars
    stmt_end: str = ";"
    # Naively, SQL statement body can contain all printable characters
    # (incl. whitespace), apart from ';'
    stmt_chars: str = printable.translate(str.maketrans("", "", stmt_end))

    # tok, or token, is parsing element with only single element output,
    # either by only having single element, or using Combine to squash
    # multiple elements into one. These are often combined to build functions
    # for matching.

    # 'rev' in variable names is shorthand for reversed.
    # Reversing happens for performance reasons,
    # as interesting things usually are at end of a long string.

    # %/%R%x%# per postgres bin/psql/settings.h
    # prompt1 per bin/psql/prompt.c:
    # %/ = current database
    # %R = =, ^
    # %x = nothing, *, !, ?
    # %# = #, >
    # This will match against "%R%x%# ", e.g "=> ".
    tok_rev_prompt_end: ParserElement = Combine(
            Opt(White())
            + (Literal('#') | Literal('>'))
            + Opt(Literal('*') | Literal('!') | Literal('?'), "")
            + (Literal('=') | Literal('^'))
    )

    # %/%R%x%# per postgres bin/psql/settings.h
    # prompt2 per bin/psql/prompt.c:
    # %/ = current database
    # %R = -, *, ', "; also $, (
    # %x = nothing, *, !, ?
    # %# = #, >
    # This will match against "%R%x%# ", e.g "-> ".
    # To save time, since linebreak prompts are only removed,
    # just match against a list of all possible combinations
    multiline_prompt_ends: list[str] = \
        ["-#", "*#", "\'#", "\"#", "$#", "(#",
         "->", "*>", "\'>", "\">", "$>", "(>",
         "-*#", "**#", "\'*#", "\"*#", "$*#", "(*#",
         "-*>", "**>", "\'*>", "\"*>", "$*>", "(*>",
         "-!#", "*!#", "\'!#", "\"!#", "$!#", "(!#",
         "-!>", "*!>", "\'!>", "\"!>", "$!>", "(!>",
         "-?#", "*?#", "\'?#", "\"?#", "$?#", "(?#",
         "-?>", "*?>", "\'?>", "\"?>", "$?>", "(?>"]
    # ParserElement for these would look this:
    #    tok_multiline_prompt_end: ParserElement = \
    #        Combine(
    #            (Literal('-') | Literal('*') | Literal('\'') |
    #             Literal('\"') | Literal('$') | Literal('(')) +
    #            Opt(Literal('*') | Literal('!') | Literal('?'), "") +
    #            (Literal('#') | Literal('>')))

    def __init__(self):
        """Plain constructor for PsqlParser."""
        pass

    def output_has_magical_return(self, psql: str) -> bool:
        """Check for weird Return presses.

        :param psql: Raw console output that includes terminal control codes.
        :returns: if output has a weird Return press.
        """
        # cheaper and easier to reverse & start from the end
        psql_rev: str = psql[::-1]  # slicing is fastest operation for reverse

        # Based on exploratory testing,
        # magic strings (related at least to ctrl-R use) are
        # "\r\n\x1b[?2004l\r", "\r\n\r\r\n" and "\x08\r\n".
        match_rev_magical_returns: ParserElement = (
            Literal("\r\n\x1b[?2004l\r"[::-1])
            | Literal("\r\n\r\r\n"[::-1])
            | Literal("\x08\r\n"[::-1])
        )

        has_magical_return: bool = False
        magical_return_res: Optional[ParseResults] = None

        try:
            magical_return_res = \
                match_rev_magical_returns.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if magical_return_res:
            has_magical_return = True
        return has_magical_return

    def output_has_new_prompt(self, psql: str) -> bool:
        """Detect when psql query evaluation has ended by parsing for a new prompt.

        :param psql: Raw console output that includes terminal control codes.
        :returns: if output is a fresh prompt.
        """
        # cheaper and easier to reverse & start from the end
        psql_rev: str = psql[::-1]
        has_new_prompt: bool = False
        prompt_res: Optional[ParseResults] = None

        match_rev_prompt_end: ParserElement = self.tok_rev_prompt_end

        try:
            prompt_res = match_rev_prompt_end.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if prompt_res:
            has_new_prompt = True

        return has_new_prompt

    def parse_new_prompt_and_rest(self, psql: str) -> list[str]:
        """Parse for a fresh prompt and everything preceding it into 2-length \
        list, facilitating easy message injection.

        :param psql: Raw console output that includes terminal control codes.
        :returns: a two-part list with everything before the prompt \
        line and then the prompt, to allow easy message injection.
        List is empty if no fresh prompt was found.
        """
        # cheaper and easier to reverse & start from the end
        psql_rev: str = psql[::-1]
        results: list[str] = []
        prompt_res: Optional[ParseResults] = None

        match_rev_prompt_and_then_rest: ParserElement = (
            self.tok_rev_prompt_end 
            + Word(self.prompt_chars)
            + (StringEnd()         # output may stop at end of db name,
               | (  # or continue \x1b[?.. 
                   Literal("?[\x1b")  # in this case control code parameter
                   + ...              # has already been parsed as prompt_chars
                   + StringEnd()
               )
            )
        )

        try:
            prompt_res = match_rev_prompt_and_then_rest.parse_string(psql_rev)
        except ParseException as e:
            print(e.explain())
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if prompt_res:
            res_list = prompt_res.as_list()
            if len(res_list) == 2:  # parsing stops right after database name
                results = [
                    '',
                    res_list[1][::-1] + res_list[0][::-1]
                ]
            elif len(res_list) == 4:  # results include \x1b[?2004h..
                results = [
                    res_list[3][::-1],
                    res_list[2][::-1] + res_list[1][::-1] + res_list[0][::-1]
                ]

        return results

    def parse_last_stmt(self, psql: str) -> str:
        """Parse for last SQL query statement in a string.

        :param psql: screenscraped psql string with only whitespace \
        after most recent query.
        :returns: parsed SQL query as plain string.
        """
        # cheaper and easier to reverse & start from the end
        psql_rev = psql[::-1]

        # Match statement that might have \r\n or whitespace at the end.
        # Parse prompt text at the end, so multiline queries can be cleaned
        # properly.
        tok_stmt_end: ParserElement = Char(';')

        match_rev_any_sql_stmt: ParserElement = (
            ZeroOrMore(White())
            + tok_stmt_end
            + ...
            + self.tok_rev_prompt_end
        )

        match_rev_last_stmt: ParserElement = (
            ZeroOrMore(Char("\n") | Char("\r") | White())
            + match_rev_any_sql_stmt
            + Word(self.prompt_chars)
        )

        results: list[str] = []
        stmt_res: Optional[ParseResults] = None
        db_name: str = ""

        try:
            stmt_res = match_rev_last_stmt.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        # If parsing was successful, pick interesting parts.
        if stmt_res is not None:
            stmt_res_list = stmt_res.as_list()
            length: int = len(stmt_res_list)

            results = [
                stmt_res_list[length - 3],
                stmt_res_list[length - 4]
            ]
            db_name = stmt_res_list[length - 1][::-1]

        # reverse back, concatenate, and remove \n's
        unreversed_flattened_res: str = reduce(
            lambda x, y: x + y[::-1], results, ""
        )

        # Replacing \n's has some edge cases where wrapper transparency
        # breaks, because both of these work right in straight psql. See:
        # "=> SELECT * FROM orders WHERE order_tot
        # al_eur = 100;"
        # should convert \n -> "" to avoid "order_tot al_eur"
        # "=> SELECT
        # * FROM orders WHERE order_total_eur = 100;"
        # should convert \n -> " " to avoid "SELECT* FROM ..".
        #
        # Replacing \n's with " " seems to have less edge cases.
        no_newlines_res = unreversed_flattened_res.replace('\n', ' ')

        # Is the statement a SELECT statement?
        match_select_stmt: ParserElement = (
            ZeroOrMore(White())
            + CaselessLiteral("SELECT")
        )
        is_select: bool = False
        try:
            is_select = \
                match_select_stmt.parse_string(no_newlines_res) is not []
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        # If it is SELECT, remove multiline delimiters and then statement is
        # ready for analysis.
        if is_select:
            demultilined_res: str = no_newlines_res
            for multiline_prompt_end in self.multiline_prompt_ends:
                prompt = db_name + multiline_prompt_end
                demultilined_res = demultilined_res.replace(prompt, "")
            return demultilined_res
        else:
            return ""

    def parse_psql_version(self, psql: str) -> str:
        """Parse for psql version and return version number.

        :param psql: psql --version output
        :returns: version string (e.g "14.5")
        """
        match_version_stmt: ParserElement = (
            Literal("psql (PostgreSQL) ")
            + Combine(
                Word(nums)
                + '.'
                + Word(nums)
            )
        )
        stmt_res: Optional[ParseResults] = None
        result: str = ""

        try:
            stmt_res = match_version_stmt.parse_string(psql)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if stmt_res:
            result = stmt_res.as_list()[1]

        return result

    def parse_syntax_error(self, psql: str) -> str:
        """Parse for syntax error output.

        :param psql: screen-scraped psql output.
        :returns: syntax error message from 'ERROR:' to last '^'.
        """
        psql_rev = psql[::-1]

        tok_marker_caret: ParserElement = Literal("^")
        tok_rev_error: ParserElement = Literal(":RORRE")
        match_error_statement: ParserElement = (
            ...
            + tok_marker_caret
            + ...
            + tok_rev_error
        )

        results: list[str] = []
        stmt_res: Optional[ParseResults] = None

        try:
            stmt_res = match_error_statement.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if stmt_res is not None:
            stmt_res_list = stmt_res.as_list()
            results = [stmt_res_list[3],
                       stmt_res_list[2],
                       stmt_res_list[1]]
        unreversed_flattened_res: str = \
            reduce(lambda x, y: x + y[::-1], results, "")

        return unreversed_flattened_res



from typing import TypedDict


# Contains all the key-value pairs meaningful in a config file.
class ConfigValues(TypedDict):
    CmpDomain: bool
    EqWildcard: bool
    ImpliedExpression: bool
    InconsistentExpression: bool
    StrangeHaving: bool
    SubqueryOrderBy: bool
    SubquerySelect: bool
    SumDistinct: bool



from typing import Optional


class ErrorFormatter:
    def __init__(
        self,
        warning_msg: str,
        warning_name: str,
        underlined_query: Optional[str] = None,
    ):
        self.warning_name: str = warning_name
        self.warning_msg: str = warning_msg
        self.underlined_query: Optional[str] = underlined_query

    def format(self) -> str:
        """
        Returns a formatted error message.
        """

        base_msg = f"Warning: {self.warning_msg} [pg4n::{self.warning_name}]"
        if self.underlined_query:
            return base_msg + f"\n{self.underlined_query}"
        return base_msg



import re
import sys
from dataclasses import dataclass
from typing import Optional

import psycopg
import sqlglot
import sqlglot.expressions as exp
from sqlglot.dialects.postgres import Postgres


@dataclass(frozen=True)
class PostgreSQLDataType:
    name: str
    digits: Optional[int]
    precision: Optional[int]


@dataclass(frozen=True)
class Column:
    name: str
    type: PostgreSQLDataType
    table: str


class SqlParser:
    # Patches the postgres dialect to recognize bpchar
    Postgres.Tokenizer.KEYWORDS["BPCHAR"] = sqlglot.TokenType.CHAR

    def __init__(self, db_connection: psycopg.Connection):
        self.dialect: str = "postgres"
        self.db_connection: psycopg.Connection = db_connection

    def parse(self, sql: str) -> list[sqlglot.exp.Expression]:
        """
        Parses all the statements in 'sql'.

        'sql' should be a string of one or more postgresql statements
        (delimited by ';'). The last ';' in 'sql' is optional.
        Raises whichever Exception sqlglot wants to throw on invalid sql.
        """

        return sqlglot.parse(sql, read=self.dialect)

    def parse_one(self, sql: str) -> sqlglot.exp.Expression:
        """
        Parses the first statement in 'sql'.
        Does not validate that 'sql' contains only 1 statement!
        'sql' should be a postgresql statement.
        The trailing ';' in 'sql' is optional.
        Raises whichever Exception sqlglot wants to throw on invalid sql.
        """

        return sqlglot.parse_one(sql, read=self.dialect)

    def get_query_columns(self, parsed_sql: exp.Expression) -> list[Column]:
        """
        Gets all columns from all tables mentioned in parsed_sql.
        """

        table_names = self.find_all_table_names(parsed_sql)

        columns = []
        for table_name in table_names:
            table_columns = self._get_columns(table_name)
            columns.extend(table_columns)

        return columns

    @staticmethod
    def get_root_node(node: exp.Expression) -> exp.Expression:
        """
        Finds the root node (probably exp.Select) from node.
        """
        root_node = node

        while True:
            new_root = root_node.find_ancestor(exp.Expression)
            if new_root is None:
                break
            root_node = new_root
        return root_node

    @staticmethod
    def find_all_table_names(parsed_sql: exp.Expression) -> list[str]:
        """
        Finds all unique table names in 'parsed_sql'.
        Any table name with an alias is listed without the alias
        (e.g.: FROM <table> AS <alias>)
        Also removes duplicate table names from the result.
        """
        table_names = []
        tables = parsed_sql.find_all(exp.Table)
        # table.this.this: first .this gets the exp.Identifier of the table and
        # second .this gets the name of the identifier.
        table_names = [table.this.this for table in tables]

        unique_table_names = list(set(table_names))

        return unique_table_names

    @staticmethod
    def get_column_name_from_column_expression(column_expression: exp.Column) -> str:
        """
        Returns the column name of column expression ast node.
        """
        return column_expression.find(exp.Identifier).this

    @staticmethod
    def find_where_predicates(root: exp.Expression) -> list[exp.Predicate]:
        """
        Finds all Predicate nodes inside Where statements from some root node
        (usually the whole parsed output from sqlparser.parse_one()).
        This function takes care to not introduce any duplicate Predicate's in
        case of nested Where statements.
        """
        predicates = []

        where_statements = root.find_all(exp.Where)

        toplevel_where_statements = list(
            filter(lambda x: not x.find_ancestor(exp.Where), where_statements)
        )

        for where_statement in toplevel_where_statements:
            for predicate in where_statement.find_all(exp.Predicate):
                predicates.append(predicate)

        return predicates

    def _get_columns(self, table_name: str) -> list[Column]:
        """
        Gets the columns from db 'db_name' table 'table_name'.
        """
        names = self._get_column_names(table_name)
        types = self._get_column_types(table_name, names)

        assert len(names) == len(types)

        columns = []
        i = 0
        for name in names:
            columns.append(Column(name, types[i], table_name))
            i += 1

        return columns

    def _get_column_names(self, table_name: str) -> list[str]:
        """
        Requires that 'table_name' exists.
        Extracts columns of the table by running postgres specific metadata
        query.
        """

        statement = (
            "SELECT column_name "
            "FROM information_schema.columns "
            f"WHERE table_name = '{table_name}';"
        )

        column_names = []
        with self.db_connection.cursor() as cursor:
            cursor.execute(statement)
            column_names = cursor.fetchall()
            self.db_connection.rollback()

        # transforms 1-element tuples to just list of elements
        column_names = [x[0] for x in column_names]

        return column_names

    def _get_column_types(
        self, table_name: str, columns: list[str]
    ) -> list[PostgreSQLDataType]:
        # sql adapted from: https://stackoverflow.com/a/20194807/13540518
        column_type_statement_prefix = f"""
SELECT
    pg_catalog.format_type(a.atttypid, a.atttypmod) AS "data_type"
FROM
    pg_catalog.pg_attribute AS a
WHERE
    a.attnum > 0
    AND NOT a.attisdropped
    AND a.attrelid = (
        SELECT c.oid
            FROM pg_catalog.pg_class c
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE c.relname = '{table_name}'
                AND pg_catalog.pg_table_is_visible(c.oid)
                AND a.attname IN ("""

        if len(columns) == 0:
            return ""

        in_condition = ""
        for col in columns[:-1]:
            in_condition += f"'{col}', "
        in_condition += f"'{columns[-1]}')\n    );"
        column_type_statement = column_type_statement_prefix + in_condition

        type_names = []
        with self.db_connection.cursor() as cursor:
            cursor.execute(column_type_statement)
            type_names = cursor.fetchall()
            self.db_connection.rollback()

        # transforms 1-element tuples to just list of elements
        type_names = [x[0] for x in type_names]

        parseable_type_names = self._convert_from_internal_types(type_names)

        return parseable_type_names

    def _convert_from_internal_types(
        self, type_names: list[str]
    ) -> list[PostgreSQLDataType]:
        """
        Postgresql has internal type names such as: character or character(x).
        sqlglot can only parse the types in the form they are declared (atleast
        in column definitions inside CREATE TABLE) like: CHAR or CHAR(x).
        This function converts from the internal to the declared form.
        """

        converted_types = []

        #  +----------------------------------------------------------+
        #  | TRIVIAL_MATCHERS                                         |
        #  +----------------------------------------------------------+
        int_matcher = re.compile(r"^(?:int|integer|int4)$", re.IGNORECASE)
        smallint_matcher = re.compile(r"^(?:smallint|int2)$", re.IGNORECASE)
        bigint_matcher = re.compile(r"^(?:bigint|int8)$", re.IGNORECASE)
        bool_matcher = re.compile(r"^(?:boolean|bool)$")
        real_matcher = re.compile(r"^(?:real|float4)$")
        bytea_matcher = re.compile(r"^(?:bytea)$")
        money_matcher = re.compile(r"^(?:money)$")
        date_matcher = re.compile(r"^(?:date)$")
        cidr_matcher = re.compile(r"^(?:cidr)$")
        inet_matcher = re.compile(r"^(?:inet)$")
        json_matcher = re.compile(r"^(?:json)$")
        jsonb_matcher = re.compile(r"^(?:jsonb)$")
        macaddr_matcher = re.compile(r"^(?:macaddr)$")
        macaddr8_matcher = re.compile(r"^(?:macaddr8)$")
        double_precission_matcher = re.compile(r"^(?:(?:double precision)|float8)$")
        text_matcher = re.compile(r"^(?:text)$")
        tsquery_matcher = re.compile(r"^(?:tsquery)$")
        tsvector_matcher = re.compile(r"^(?:tsvector)$")
        uuid_matcher = re.compile(r"^(?:uuid)$")
        xml_matcher = re.compile(r"^(?:xml)$")
        circle_matcher = re.compile(r"^(?:circle)$")
        box_matcher = re.compile(r"^(?:box)$")
        path_matcher = re.compile(r"^(?:path)$")
        line_matcher = re.compile(r"^(?:line)$")
        lseg_matcher = re.compile(r"^(?:lseg)$")
        point_matcher = re.compile(r"^(?:point)$")
        polygon_matcher = re.compile(r"^(?:polygon)$")
        interval_matcher = re.compile(r"^(?:interval)$")

        #  +----------------------------------------------------------+
        #  | OPT_PRECISSION_MATCHERS                                  |
        #  +----------------------------------------------------------+
        character_matcher = re.compile(r"^(?:character|char)(?:\(\s*(\d+)\s*\))?$")
        bit_matcher = re.compile(r"^(?:bit)(?:\(\s*(\d+)\s*\))?$")
        varchar_matcher = re.compile(
            r"^(?:(?:character varying)|varchar)(?:\(\s*(\d+)\s*\))?$"
        )
        varbit_matcher = re.compile(r"^(?:(?:bit varying)|varbit)(?:\(\s*(\d+)\s*\))?$")

        timestamp_matcher = re.compile(
            r"^timestamp(?:\(\s*(\d+)\s*\))?(?: without time zone)?$"
        )
        timestamptz_matcher = re.compile(
            r"^(?:timestamp|timestamptz)(?:\(\s*(\d+)\s*\))?(?: with time zone)?$"
        )
        time_matcher = re.compile(r"^time(?:\(\s*(\d+)\s*\))?(?: without time zone)?$")
        timetz_matcher = re.compile(
            r"^(?:time|timetz)(?:\(\s*(\d+)\s*\))?(?: with time zone)?$"
        )

        #  +----------------------------------------------------------+
        #  | COMPLEX_MATCHERS                                         |
        #  +----------------------------------------------------------+
        numeric_matcher = re.compile(
            r"^(?:numeric|decimal)(?:\(\s*(\d+)(?:\s*,\s*(\d+))?\s*\))?$"
        )

        for type_name in type_names:
            #  +----------------------------------------------------------+
            #  | TRIVIAL_CASES                                            |
            #  +----------------------------------------------------------+
            if match := int_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("INT", None, None))
            elif match := smallint_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("SMALLINT", None, None))
            elif match := bigint_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("BIGINT", None, None))
            elif match := bool_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("BOOL", None, None))
            elif match := real_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("REAL", None, None))
            elif match := bytea_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("BYTEA", None, None))
            elif match := money_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("MONEY", None, None))
            elif match := date_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("DATE", None, None))
            elif match := cidr_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("CIDR", None, None))
            elif match := inet_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("INET", None, None))
            elif match := json_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("JSON", None, None))
            elif match := jsonb_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("JSONB", None, None))
            elif match := macaddr_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("MACADDR", None, None))
            elif match := macaddr8_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("MACADDR8", None, None))
            elif match := double_precission_matcher.match(type_name):
                converted_types.append(
                    PostgreSQLDataType("DOUBLE_PRECISSION", None, None)
                )
            elif match := text_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("TEXT", None, None))
            elif match := tsquery_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("TSQUERY", None, None))
            elif match := tsvector_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("TSVECTOR", None, None))
            elif match := uuid_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("UUID", None, None))
            elif match := xml_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("XML", None, None))
            elif match := circle_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("CIRCLE", None, None))
            elif match := box_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("BOX", None, None))
            elif match := path_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("PATH", None, None))
            elif match := line_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("LINE", None, None))
            elif match := lseg_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("LSEG", None, None))
            elif match := point_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("POINT", None, None))
            elif match := polygon_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("POLYGON", None, None))
            elif match := interval_matcher.match(type_name):
                converted_types.append(PostgreSQLDataType("INTERVAL", None, None))

            #  +----------------------------------------------------------+
            #  | OPT_PRECISSION_CASES                                     |
            #  +----------------------------------------------------------+
            elif match := character_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "CHAR")
                )
            elif match := bit_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "BIT")
                )
            elif match := varchar_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "VARCHAR")
                )
            elif match := varbit_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "VARBIT")
                )
            elif match := timestamp_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "TIMESTAMP")
                )
            elif match := timestamptz_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "TIMESTAMPTZ")
                )
            elif match := time_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "TIME")
                )
            elif match := timetz_matcher.match(type_name):
                converted_types.append(
                    self._eval_opt_precission_field_match(match, "TIMETZ")
                )

            #  +----------------------------------------------------------+
            #  | COMPLEX_CASES                                            |
            #  +----------------------------------------------------------+
            elif match := numeric_matcher.match(type_name):
                if match.group(1) is None and match.group(2) is None:
                    conv_type = PostgreSQLDataType("DECIMAL", None, None)
                    converted_types.append(conv_type)
                elif match.group(1) is not None and match.group(2) is None:
                    conv_type = PostgreSQLDataType(
                        f"DECIMAL({match.group(1)})", int(match.group(1)), None
                    )
                elif match.group(1) is not None and match.group(2) is not None:
                    name = f"DECIMAL({match.group(1)},{match.group(2)})"
                    conv_type = PostgreSQLDataType(
                        name, int(match.group(1)), int(match.group(2))
                    )
                    converted_types.append(conv_type)
                else:
                    print(
                        f"unrecognized number type format '{type_name}' for numeric() column type",
                        file=sys.stderr,
                    )
                    raise Exception
            else:
                print(
                    f"unable to convert from internal type '{type_name}' to declared type",
                    file=sys.stderr,
                )
                raise Exception

        return converted_types

    def _eval_opt_precission_field_match(
        self, match: re.Match, type_name_prefix: str
    ) -> PostgreSQLDataType:
        """
        Evaluates a internal type of the form: '<typename>[ (p) ]'
        where p is optional precission field and returns a converted type used
        "normalized" in the context of this program's conventions.
        """
        if match.group(1) is None:
            return PostgreSQLDataType("TIMESTAMPTZ", None, None)
        return PostgreSQLDataType(
            f"{type_name_prefix}({match.group(1)})", int(match.group(1)), None
        )



"""Test PsqlParser."""

from ..psqlparser import PsqlParser


def test_output_has_new_prompt() -> None:
    p = PsqlParser()

    case_trivial = \
        "\x1b[?2004hpgdb=> "
    assert p.output_has_new_prompt(case_trivial)

    case_mixed_results = \
        "\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\x1b[?2004hpgdb=# "
    assert p.output_has_new_prompt(case_mixed_results)


def test_parse_new_prompt_and_rest() -> None:
    p = PsqlParser()

    case_trivial = \
        "\x1b[?2004hpgdb=> "
    assert p.parse_new_prompt_and_rest(case_trivial) == \
        ["", "\x1b[?2004hpgdb=> "]

    case_mixed_results = \
        "\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\x1b[?2004hpgdb=# "
    assert p.parse_new_prompt_and_rest(case_mixed_results) == \
        ["\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n", "\x1b[?2004hpgdb=# "]


def test_output_has_magical_return() -> None:
    p = PsqlParser()

    case_long_return = \
        "\r\x1b[16Ppgdb=# SELECT * FROM orders  WHERE order_total_eur = 0 AND order_tot\x08\r\n\x1b[?2004l\r"
    assert p.output_has_magical_return(case_long_return)

def test_parse_last_stmt() -> None:
    p = PsqlParser()

    case_trivial = \
        "psql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM orders;"
    assert p.parse_last_stmt(case_trivial) == \
        "SELECT * FROM orders;"

    case_select_then_select = \
        "psql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM orders;\npgdb=# SELECT * FROM orders WHERE order_total_eur = 0 AND order_total_eur = 100;"
    assert p.parse_last_stmt(case_select_then_select) == \
        "SELECT * FROM orders WHERE order_total_eur = 0 AND order_total_eur = 100;"

    # Broken case found in empirical testing
    case_empirical_1 = \
        "sql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM orders  WHERE order_total_eur = 0 AND order_total_eur = 100;\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
    assert p.parse_last_stmt(case_empirical_1) == \
        "SELECT * FROM orders  WHERE order_total_eur = 0 AND order_total_eur = 100;"

    case_query_with_newline = \
        "sql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM orders WHERE\norder_total_eur = 100;\n\n\n"
    assert p.parse_last_stmt(case_query_with_newline) == \
        "SELECT * FROM orders WHERE order_total_eur = 100;"
    # current unfixed bug case, needs to be fixed after mid-presentations:
    case_select_then_insert = "psql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM orders;\npgdb=# INSERT INTO orders VALUES (6, 6, 6);"
    assert p.parse_last_stmt(case_select_then_insert) == \
        ""

    case_multiple_queries_and_whitespaces = \
        "psql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM orders;\npgdb=# INSERT INTO orders VALUES (6, 6, 6);\npgdb=#   SELECT    * FROM\n  orders    WHERE order_total_eur = 100   ; "
    assert p.parse_last_stmt(case_multiple_queries_and_whitespaces) == \
        "SELECT    * FROM   orders    WHERE order_total_eur = 100   ;"

    case_multiline_query = \
        "psql (14.5)\nType \"help\" for help.\n\npgdb=# SELECT * FROM\npgdb-# orders;"
    assert p.parse_last_stmt(case_multiline_query) == \
        "SELECT * FROM  orders;"



from functools import reduce
import sys
from typing import Optional

import pexpect

from .psqlconninfo import PsqlConnInfo
from .semanticrouter import SemanticRouter
from .psqlparser import PsqlParser
from .psqlwrapper import PsqlWrapper
from .config_reader import ConfigReader
from .config_values import ConfigValues


def main() -> None:
    """Initiate session by getting psql connection parameters via psql \
    child process, initializing semantic analysis and wrapper modules, \
    and then starting the session."""

    # Configuration is ignored if there is error reading config files
    config_values: Optional[ConfigValues] = None
    try:
        config_reader = ConfigReader()
        config_values = config_reader.read()
    except Exception as e:
        config_values = None

    if len(sys.argv) > 1:
        conn_info = PsqlConnInfo(
            reduce(lambda x, y: x + y, sys.argv[1:], "")  # concat arguments
            ).get()
        if conn_info is not None:
            # asterisk unpacks the 5-tuple
            sem_router = SemanticRouter(*conn_info, config_values)
            psql = PsqlWrapper(
                sys.argv[1].encode("utf-8"),
                # semantic analysis:
                sem_router.run_analysis,
                # no syntax error analysis:
                lambda syntax_error_analysis: "",
                PsqlParser()
            )
            psql.start()
        else:
            # Psql is not connecting to any database,
            # e.g. "pg4n --help" is being run.
            # for simplicity, just use pexpect here
            psql_output = pexpect.spawn(
                "psql "
                + reduce(
                    lambda x, y: x + y, sys.argv[1:], ""
                )
            )
            psql_output.expect(pexpect.EOF)
            print(bytes.decode(psql_output.before))
    else:
        print("pg4n [psql arguments] <database name>")


if __name__ == "__main__":
    main()



from tempfile import TemporaryFile
from typing import Optional

from ..config_parser import ConfigParser
from ..config_values import ConfigValues


def test_parse():
    CONFIG = """ImpliedExpression 0

######################
 # # #
######   # # #
              ##       3              ## jkelm el el e
CmpDomain false
SubqueryOrderBy true
ImpliedExpression yes

SUBQUERYORDERBY true
subqueryorderby true


SuBqUeRyOrDeRbY true

# Gives too many false positives
SubquerySelect 0
SubquerySelect 0
#
"""

    with TemporaryFile(buffering=0) as tmp_file:
        tmp_file.write(bytes(CONFIG, "utf-8"))
        tmp_file.flush()
        tmp_file.seek(0)

        parser = ConfigParser(tmp_file)
        config_values: Optional[ConfigValues] = parser.parse()
        assert config_values is not None

        try:
            from pprint import pprint
            pprint(config_values)
            assert config_values["ImpliedExpression"] == True
            assert config_values["CmpDomain"] == False
            assert config_values["SubqueryOrderBy"] == True
            assert config_values["SubquerySelect"] == False
        except Exception as e:
            assert False, f"{e}"


# TODO: The actual test
# def test_multiple_option_definition_warings():

# This is the config file
#     CONFIG1 = """ImpliedExpression false
# ImpliedExpression 1
# CmpDomain false
# SubqueryOrderBy true
# SubqueryOrderBy true
# SubqueryOrderBy true
# SubqueryOrderBy true
# SubqueryOrderBy true

# SubqueryOrderBy true
# SubquerySelect yes
# """

# This is the matching output of running the program that reads only 1 config
# file CONFIG
# configuration file output is not tested because its system dependent
# warning_matcher = re.compile(\
# """
# line 4: 'SubqueryOrderBy true'
# line 5: 'SubqueryOrderBy true'
# line 6: 'SubqueryOrderBy true'
# line 7: 'SubqueryOrderBy true'
# line 8: 'SubqueryOrderBy true'
# line 10: 'SubqueryOrderBy true'
# in configuration file: '.*'
# warning: option 'ImpliedExpression' is set multiple times
# line 1: 'ImpliedExpression false'
# line 2: 'ImpliedExpression 1'
# in configuration file: '.*'
# )



from typing import Optional

import sqlglot.expressions as exp

from .qepparser import QEPAnalysis
from .errfmt import ErrorFormatter


class StrangeHavingChecker:
    def __init__(self, parsed_sql: exp.Expression, qep_analysis: QEPAnalysis):
        self.parsed_sql: exp.Expression = parsed_sql
        self.qep_analysis: QEPAnalysis = qep_analysis

    def check(self) -> Optional[str]:
        """
        Returns warning message if there exists HAVING without a GROUP BY,
        otherwise None.
        """

        has_group_by = len(list(self.parsed_sql.find_all(exp.Group))) != 0
        has_having = len(list(self.parsed_sql.find_all(exp.Having))) != 0
        has_strange_having = has_having and not has_group_by

        if not has_strange_having:
            return None

        warning = "HAVING without GROUP BY"
        warning_name = type(self).__name__.rstrip("Checker")
        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg



from typing import Optional

import sqlglot.expressions as exp

from .errfmt import ErrorFormatter
from .qepparser import QEPAnalysis, QEPNode


class InconsistentExpressionChecker:
    """
    Inconsistent expression is some expression that is never true.
    For example: x = 10 AND x = 20

    This checker only finds a small subset of such expression, where postgresql
    itself detects the inconsistent expression in its query optimizer and
    exposes that information via its query execution plan.

    """

    # TODO: Analyze the where expressions of the sql and find more inconsistent
    #       expressions.
    #       - One approach is to transform the expressions with its symbols from
    #         sqlparser representation into a formula and finding whether it is
    #         satisfiable with pysmt.
    #       - This approach could also be used to find tautologies.
    #       - Some preliminary work of this is in the experimental branch
    #         feat/experimental-smt

    def __init__(self, parsed_sql: exp.Expression, qep_analysis: QEPAnalysis):
        self.parsed_sql: exp.Expression = parsed_sql
        self.qep_analysis: QEPAnalysis = qep_analysis

    def check(self) -> Optional[str]:
        """
        Returns warning msg if the sql contains inconsistent expression,
        otherwise None.
        """

        if self.qep_analysis is None:
            return None

        def finder(node: QEPNode) -> bool:
            return node.get("One-Time Filter") is not None

        has_onetime_filter = len(self.qep_analysis.root.rfind(finder)) > 0
        has_inconsistent_expression = has_onetime_filter

        if not has_inconsistent_expression:
            return None

        warning_name = type(self).__name__.rstrip("Checker")
        warning = "Some condition is always false"
        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg



from itertools import chain
from typing import Callable, Iterable, List, TypedDict
import psycopg
from psycopg import Connection

from . import util  # used to test relative imports


# TODO: break into variants discriminated by Node Type
# right now, the interface isn't safe to use because it's not clear what fields
# are available for each node type
node = TypedDict("Plan", {
    "Node Type": str,
    "Parent Relationship": str,
    "Startup Cost": float,
    "Total Cost": float,
    "Plan Rows": int,
    "Plan Width": int,
    "Actual Startup Time": float,
    "Actual Total Time": float,
    "Actual Rows": int,
    "Actual Loops": int,
    "Plans": List["node"],
    "Index Cond": str,
    "Filter": str,
    "Relation Name": str,
    "Alias": str,
    "Scan Direction": str,
    "Index Name": str,
    "Triggers": list[str],
    "Total Runtime": float,
    "One-Time Filter": str,
})

qep = TypedDict("QEP", {
    "Plan": node,
    "Triggers": list[str],
    "Planning Time": float,
    "Execution Time": float,
    "Total Runtime": float,
})


class QEPNode:
    """A node in a query execution plan."""

    def __init__(self, node_: node):
        """Create a new QEPNode.

        :param node_: the node to wrap"""
        self._node = node_

    def __iter__(self) -> Iterable["QEPNode"]:
        """Iterate over child nodes."""
        return map(QEPNode, self._node.get("Plans", []))

    def __len__(self) -> int:
        """Get the number of child nodes."""
        return len(self._node["Plans"])

    def __getitem__(self, key: int) -> "QEPNode":
        """Get the child node at the given index.

        :param key: the index of the child node to get

        :returns: the child node at the given index
        """
        return QEPNode(self._node["Plans"][key])

    def __str__(self):
        return self._node.__str__()

    def __repr__(self):
        return self._node.__repr__()

    @property
    def plan(self) -> node:
        """A dict of the node's properties."""
        return self._node

    @property
    def plans(self) -> list[node]:
        """A list of the node's children."""
        return self._node.get("Plans", [])

    def find(self, pr: Callable[[node], bool],
             recursive=False) -> list[node]:
        """Finds nodes matching the predicate.

        :param pr: a function that takes a node and returns True if it matches
        :param recursive: if True, search recursively, otherwise only search
            this+children

        :returns: a list of matching nodes
        """
        if recursive:
            return self.find(pr) + \
                list(chain.from_iterable(x.find(pr, True) for x in iter(self)))
        return list(filter(pr, chain((self._node,), self.plans)))

    def rfind(self, pred: Callable[[node], bool]) -> list[node]:
        """Finds nodes matching the predicate, recursively.

        :param pred: a function that takes a node and returns True if it matches

        :returns: a list of matching nodes
        """
        return self.find(pred, recursive=True)

    def findval(self, key: str, val: object, recursive=False) -> list[node]:
        """Finds nodes with the given key and value.

        :param key: the key to search for
        :param val: the value to search for
        :param recursive: if True, search recursively, otherwise only search
            this+children

        :returns: a list of matching nodes
        """
        return self.find(lambda x: x.get(key) == val, recursive)

    def rfindval(self, key: str, val: object) -> list[node]:
        """Finds nodes with the given key and value, recursively.

        :param key: the key to search for
        :param val: the value to search for

        :returns: a list of matching nodes
        """
        return self.findval(key, val, recursive=True)


class QEPAnalysis:
    """Represents the result of EXPLAIN ANALYZE."""

    def __init__(self, qep_: qep):
        self._qep = qep_

    def __str__(self):
        return self._qep.__str__()

    def __repr__(self):
        return self._qep.__repr__()

    @property
    def root(self) -> QEPNode:
        """The root node of the query execution plan."""
        return QEPNode(self._qep["Plan"])

    @property
    def plan(self) -> node:
        """A dict of the root node's properties."""
        return self._qep["Plan"]

    @property
    def qep(self) -> qep:
        """A dict of the query execution plan's properties."""
        return self._qep


class QEPParser:
    """Performs analyses on given queries, returning resultant QEPAnalysis."""

    def __init__(self, *args, conn=None, constraint_exclusion=True, **kwargs):
        self._ref = bool(conn)
        self._conn: Connection = conn or psycopg.connect(*args, **kwargs)
        # use constraint_exclusion to avoid unnecessary index scans
        if constraint_exclusion:
            with self._conn.cursor() as cur:
                try:
                    cur.execute("set constraint_exclusion = on;")
                    self._conn.commit()
                except Exception as e:
                    self._conn.rollback()
        else:
            with self._conn.cursor() as cur:
                try:
                    cur.execute("set constraint_exclusion = off;")
                    self._conn.commit()
                except Exception as e:
                    self._conn.rollback()

    def __del__(self):
        if not self._ref:
            self._conn.close()

    def __call__(self, stmt: str, *args, **kwargs) -> QEPAnalysis:
        """
        Executes a query and returns the query execution plan as a dictionary.

        Parameters:
            stmt: The query to execute.
            *args: Positional arguments to pass to cursor.execute().
            **kwargs: Keyword arguments to pass to cursor.execute().

        Returns:
            A dictionary representing the query execution plan.
        """
        stmt = "explain (format json, analyze, verbose) " + \
            stmt.strip().rstrip(';') + ";"
        try:
            with self._conn.cursor() as cur:
                cur.execute(stmt, *args, **kwargs)
                res = cur.fetchall()
                self._conn.rollback()
                if (n := len(res)) != 1:
                    raise ValueError(f"Expected 1 row, got {n}")
                if (n := len(res[0])) != 1:
                    raise ValueError(f"Expected 1 column, got {n}")
                if (n := len(res[0][0])) != 1:
                    raise ValueError(f"Expected 1 item in column, got {n}")
                if (t := type(res[0][0][0])) != dict:
                    raise ValueError(f"Expected dict in column, got {t}")
                return QEPAnalysis(res[0][0][0])
        except psycopg.Error as e:
            self._conn.rollback()


    def parse(self, stmt: str, *args, **kwargs) -> QEPAnalysis:
        '''Alias for __call__'''
        return self(stmt, *args, **kwargs)
